# -*- coding: utf-8 -*-

# spechomo, Spectral homogenization of multispectral satellite data
#
# Copyright (C) 2019-2021
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Please note the following exception: `spechomo` depends on tqdm, which is
# distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import re
from collections import OrderedDict
from typing import Union, List  # noqa F401  # flake8 issue
from tqdm import tqdm

import numpy as np
from geoarray import GeoArray
from pandas import DataFrame
from pandas.plotting import scatter_matrix
from pyrsr import RSR

from .utils import im2spectra


class TrainingData(object):
    """Class for analyzing statistical relations between a pair of machine learning training data cubes."""

    def __init__(self, im_X, im_Y, test_size):
        # type: (Union[GeoArray, np.ndarray], Union[GeoArray, np.ndarray], Union[float, int]) -> None
        """Get instance of TrainingData.

        :param im_X:        input image X
        :param im_Y:        input image Y
        :param test_size:   test size (proportion as float between 0 and 1) or absolute value as integer
        """
        from sklearn.model_selection import train_test_split  # avoids static TLS error here

        self.im_X = GeoArray(im_X)
        self.im_Y = GeoArray(im_Y)

        # Set spectra (3D to 2D conversion)
        self.spectra_X = im2spectra(self.im_X)
        self.spectra_Y = im2spectra(self.im_Y)

        # Set train and test variables
        # NOTE: If random_state is set to an Integer, train_test_split will always select the same 'pseudo-random' set
        #       of the input data.
        self.train_X, self.test_X, self.train_Y, self.test_Y = \
            train_test_split(self.spectra_X, self.spectra_Y, test_size=test_size, shuffle=True, random_state=0)

    def plot_scatter_matrix(self, figsize=(15, 15), mode='intersensor'):
        # TODO complete this function
        from matplotlib import pyplot as plt

        train_X = self.train_X[np.random.choice(self.train_X.shape[0], 1000, replace=False), :]
        train_Y = self.train_Y[np.random.choice(self.train_Y.shape[0], 1000, replace=False), :]

        if mode == 'intersensor':
            import seaborn

            fig, axes = plt.subplots(train_X.shape[1], train_Y.shape[1],
                                     figsize=(25, 9), sharex='all', sharey='all')
            # fig.suptitle('Correlation of %s and %s bands' % (self.src_cube.satellite, self.tgt_cube.satellite),
            #              size=25)

            color = seaborn.hls_palette(13)

            for i, ax in zip(range(train_X.shape[1]), axes.flatten()):
                for j, ax in zip(range(train_Y.shape[1]), axes.flatten()):
                    axes[i, j].scatter(train_X[:, j], train_Y[:, i], c=color[j], label=str(j))
                    # axes[i, j].set_xlim(-0.1, 1.1)
                    # axes[i, j].set_ylim(-0.1, 1.1)
                    #  if j == 8:
                    #      axes[5, j].set_xlabel('S2 B8A\n' + str(metadata_s2['Bands_S2'][j]) + ' nm', size=10)
                    #  elif j in range(9, 13):
                    #      axes[5, j].set_xlabel('S2 B' + str(j) + '\n' + str(metadata_s2['Bands_S2'][j]) + ' nm',
                    #                            size=10)
                    #  else:
                    #      axes[5, j].set_xlabel('S2 B' + str(j + 1) + '\n' + str(metadata_s2['Bands_S2'][j]) + ' nm',
                    #                            size=10)
                    #  axes[i, 0].set_ylabel(
                    #      'S3 SLSTR B' + str(6 - i) + '\n' + str(metadata_s3['Bands_S3'][5 - i]) + ' nm',
                    #      size=10)
                    # axes[4, j].set_xticks(np.arange(0, 1.2, 0.2))
                    # axes[i, j].plot([0, 1], [0, 1], c='red')

        else:
            df = DataFrame(train_X, columns=['Band %s' % b for b in range(1, self.im_X.bands + 1)])
            scatter_matrix(df, figsize=figsize, marker='.', hist_kwds={'bins': 50}, s=30, alpha=0.8)
            plt.suptitle('Image X band to band correlation')

            df = DataFrame(train_Y, columns=['Band %s' % b for b in range(1, self.im_Y.bands + 1)])
            scatter_matrix(df, figsize=figsize, marker='.', hist_kwds={'bins': 50}, s=30, alpha=0.8)
            plt.suptitle('Image Y band to band correlation')

    def plot_scattermatrix(self):
        # TODO complete this function
        import seaborn
        from matplotlib import pyplot as plt

        fig, axes = plt.subplots(self.im_X.data.bands, self.im_Y.data.bands,
                                 figsize=(25, 9), sharex='all', sharey='all')
        fig.suptitle('Correlation of %s and %s bands' % (self.im_X.satellite, self.im_Y.satellite), size=25)

        color = seaborn.hls_palette(13)

        for i, ax in zip(range(6), axes.flatten()):
            for j, ax in zip(range(13), axes.flatten()):
                axes[i, j].scatter(self.train_X[:, j], self.train_Y[:, 5 - i], c=color[j], label=str(j))
                axes[i, j].set_xlim(-0.1, 1.1)
                axes[i, j].set_ylim(-0.1, 1.1)
                # if j == 8:
                #     axes[5, j].set_xlabel('S2 B8A\n' + str(metadata_s2['Bands_S2'][j]) + ' nm', size=10)
                # elif j in range(9, 13):
                #     axes[5, j].set_xlabel('S2 B' + str(j) + '\n' + str(metadata_s2['Bands_S2'][j]) + ' nm', size=10)
                # else:
                #     axes[5, j].set_xlabel('S2 B' + str(j + 1) + '\n' + str(metadata_s2['Bands_S2'][j]) + ' nm',
                #                           size=10)
                # axes[i, 0].set_ylabel('S3 SLSTR B' + str(6 - i) + '\n' + str(metadata_s3['Bands_S3'][5 - i]) + ' nm',
                #                       size=10)
                axes[4, j].set_xticks(np.arange(0, 1.2, 0.2))
                axes[i, j].plot([0, 1], [0, 1], c='red')

    def show_band_scatterplot(self, band_src_im, band_tgt_im):
        # TODO complete this function
        from scipy.stats import gaussian_kde
        from matplotlib import pyplot as plt

        x = self.im_X.data[band_src_im].flatten()[:10000]
        y = self.im_Y.data[band_tgt_im].flatten()[:10000]

        # Calculate the point density
        xy = np.vstack([x, y])
        z = gaussian_kde(xy)(xy)

        plt.figure(figsize=(15, 15))
        plt.scatter(x, y, c=z, s=30, edgecolor='none')
        plt.show()


class RefCube(object):
    """Data model class for reference cubes holding the training data for later fitted machine learning classifiers."""

    def __init__(self, filepath='', satellite='', sensor='', LayerBandsAssignment=None):
        # type: (str, str, str, list) -> None
        """Get instance of RefCube.

        :param filepath:                file path for importing an existing reference cube from disk
        :param satellite:               the satellite for which the reference cube holds its spectral data
        :param sensor:                  the sensor for which the reference cube holds its spectral data
        :param LayerBandsAssignment:    the LayerBandsAssignment for which the reference cube holds its spectral data
        """
        # privates
        self._col_imName_dict = dict()
        self._wavelenths = []

        # defaults
        self.data = GeoArray(np.empty((0, 0, len(LayerBandsAssignment) if LayerBandsAssignment else 0)),
                             nodata=-9999)
        self.srcImNames = []

        # args/ kwargs
        self.filepath = filepath
        self.satellite = satellite
        self.sensor = sensor
        self.LayerBandsAssignment = LayerBandsAssignment or []

        if filepath:
            self.read_data_from_disk(filepath)

        if self.satellite and self.sensor and self.LayerBandsAssignment:
            self._add_bandnames_wavelenghts_to_meta()

    def _add_bandnames_wavelenghts_to_meta(self):
        # set bandnames
        self.data.bandnames = ['Band %s' % b for b in self.LayerBandsAssignment]

        # set wavelengths
        self.data.metadata.band_meta['wavelength'] = self.wavelengths

    @property
    def n_images(self):
        """Return the number training images from which the reference cube contains spectral samples."""
        return self.data.shape[1]

    @property
    def n_signatures(self):
        """Return the number spectral signatures per training image included in the reference cube."""
        return self.data.shape[0]

    @property
    def n_clusters(self):
        """Return the number spectral clusters used for clustering source images for the reference cube."""
        if self.filepath:
            identifier = re.search('refcube__(.*).bsq', os.path.basename(self.filepath)).group(1)
            return int(identifier.split('__')[2].split('nclust')[1])

    @property
    def n_signatures_per_cluster(self):
        if self.n_clusters:
            return self.n_signatures // self.n_clusters

    @property
    def col_imName_dict(self):
        # type: () -> OrderedDict
        """Return an ordered dict containing the file base names of the original training images for each column."""
        return OrderedDict((col, imName) for col, imName in zip(range(self.n_images), self.srcImNames))

    @col_imName_dict.setter
    def col_imName_dict(self, col_imName_dict):
        # type: (dict) -> None
        self._col_imName_dict = col_imName_dict
        self.srcImNames = list(col_imName_dict.values())

    @property
    def wavelengths(self):
        if not self._wavelenths and self.satellite and self.sensor and self.LayerBandsAssignment:
            self._wavelenths = list(RSR(self.satellite, self.sensor,
                                        LayerBandsAssignment=self.LayerBandsAssignment).wvl)

        return self._wavelenths

    @wavelengths.setter
    def wavelengths(self, wavelengths):
        self._wavelenths = wavelengths

    def add_refcube_array(self, refcube_array, src_imnames, LayerBandsAssignment):
        # type: (Union[str, np.ndarray], list, list) -> None
        """Add the given given array to the RefCube instance.

        :param refcube_array:           3D array or file path  of the reference cube to be added
                                        (spectral samples /signatures x training images x spectral bands)
        :param src_imnames:             list of training image file base names from which the given cube received data
        :param LayerBandsAssignment:    LayerBandsAssignment of the spectral bands of the given 3D array
        :return:
        """
        # validation
        assert LayerBandsAssignment == self.LayerBandsAssignment, \
            "%s != %s" % (LayerBandsAssignment, self.LayerBandsAssignment)

        if self.data.size:
            new_cube = np.hstack([self.data, refcube_array])
            self.data = GeoArray(new_cube, nodata=self.data.nodata)
        else:
            self.data = GeoArray(refcube_array, nodata=self.data.nodata)

        self.srcImNames.extend(src_imnames)

    def add_spectra(self, spectra, src_imname, LayerBandsAssignment):
        # type: (np.ndarray, str, list) -> None
        """Add a set of spectral signatures to the reference cube.

        :param spectra:              2D numpy array with rows: spectral samples / columns: spectral information (bands)
        :param src_imname:           image basename of the source hyperspectral image
        :param LayerBandsAssignment: LayerBandsAssignment for the spectral dimension of the passed spectra,
                                     e.g., ['1', '2', '3', '4', '5', '6L', '6H', '7', '8']
        """
        # validation
        assert LayerBandsAssignment == self.LayerBandsAssignment, \
            "%s != %s" % (LayerBandsAssignment, self.LayerBandsAssignment)

        # reshape 2D spectra array to one image column (refcube is an image with spectral information in the 3rd dim.)
        im_col = spectra.reshape((spectra.shape[0], 1, spectra.shape[1]))

        meta = self.data.metadata  # needs to be copied to the new GeoArray

        if self.data.size:
            # validation
            if spectra.shape[0] != self.data.shape[0]:
                raise ValueError('The number of signatures in the given spectra array does not match the dimensions of '
                                 'the reference cube.')

            # append spectra to existing reference cube
            new_cube = np.hstack([self.data, im_col])
            self.data = GeoArray(new_cube, nodata=self.data.nodata)

        else:
            self.data = GeoArray(im_col, nodata=self.data.nodata)

        # copy previous metadata to the new GeoArray instance
        self.data.metadata = meta

        # add source image name to list of image names
        self.srcImNames.append(src_imname)

    @property
    def metadata(self):
        """Return an ordered dictionary holding the metadata of the reference cube."""
        attrs2include = ['satellite', 'sensor', 'filepath', 'n_signatures', 'n_images', 'n_clusters',
                         'n_signatures_per_cluster', 'col_imName_dict', 'LayerBandsAssignment', 'wavelengths']
        return OrderedDict((k, getattr(self, k)) for k in attrs2include)

    def get_band_combination(self, tgt_LBA):
        # type: (List[str]) -> GeoArray
        """Get an array according to the bands order given by a target LayerBandsAssignment.

        :param tgt_LBA:     target LayerBandsAssignment
        :return:
        """
        if tgt_LBA != self.LayerBandsAssignment:
            cur_LBA_dict = dict(zip(self.LayerBandsAssignment, range(len(self.LayerBandsAssignment))))
            tgt_bIdxList = [cur_LBA_dict[lr] for lr in tgt_LBA]

            return GeoArray(np.take(self.data, tgt_bIdxList, axis=2), nodata=self.data.nodata)
        else:
            return self.data

    def get_spectra_dataframe(self, tgt_LBA):
        # type: (List[str]) -> DataFrame
        """Return a pandas.DataFrame [sample x band] according to the given LayerBandsAssignment.

        :param tgt_LBA: target LayerBandsAssignment
        :return:
        """
        imdata = self.get_band_combination(tgt_LBA)
        spectra = im2spectra(imdata)
        df = DataFrame(spectra, columns=['B%s' % band for band in tgt_LBA])

        return df

    def rearrange_layers(self, tgt_LBA):
        # type: (List[str]) -> None
        """Rearrange the spectral bands of the reference cube according to the given LayerBandsAssignment.

        :param tgt_LBA:     target LayerBandsAssignment
        """
        self.data = self.get_band_combination(tgt_LBA)
        self.LayerBandsAssignment = tgt_LBA

    def read_data_from_disk(self, filepath):
        self.data = GeoArray(filepath)

        with open(os.path.splitext(filepath)[0] + '.meta', 'r') as metaF:
            meta = json.load(metaF)
            for k, v in meta.items():
                if k in ['n_signatures', 'n_images', 'n_clusters', 'n_signatures_per_cluster']:
                    continue  # skip pure getters
                else:
                    setattr(self, k, v)

    def save(self, path_out, fmt='ENVI'):
        # type: (str, str) -> None
        """Save the reference cube to disk.

        :param path_out:    output path on disk
        :param fmt:         output format as GDAL format code
        :return:
        """
        self.filepath = self.filepath or path_out
        self.data.save(out_path=path_out, fmt=fmt)

        # save metadata as JSON file
        with open(os.path.splitext(path_out)[0] + '.meta', 'w') as metaF:
            json.dump(self.metadata.copy(), metaF, separators=(',', ': '), indent=4)

    def _get_spectra_by_label_imname(self, cluster_label, image_basename, n_spectra2get=100, random_state=0):
        cluster_start_pos_all = list(range(0, self.n_signatures, self.n_signatures_per_cluster))
        cluster_start_pos = cluster_start_pos_all[cluster_label]
        spectra = self.data[cluster_start_pos: cluster_start_pos + self.n_signatures_per_cluster,
                            self.srcImNames.index(image_basename)]
        idxs_specIncl = np.random.RandomState(seed=random_state).choice(range(self.n_signatures_per_cluster),
                                                                        n_spectra2get)
        return spectra[idxs_specIncl, :]

    def plot_sample_spectra(self, image_basename, cluster_label='all', include_mean_spectrum=True,
                            include_median_spectrum=True, ncols=5, **kw_fig):
        # type: (Union[str, int, List], str, bool, bool, int, dict) -> 'plt.figure'
        from matplotlib import pyplot as plt

        if isinstance(cluster_label, int):
            lbls2plot = [cluster_label]
        elif isinstance(cluster_label, list):
            lbls2plot = cluster_label
        elif cluster_label == 'all':
            lbls2plot = list(range(self.n_clusters))
        else:
            raise ValueError(cluster_label)

        # create a single plot
        if len(lbls2plot) == 1:
            if cluster_label == 'all':
                cluster_label = 0

            fig, axes = plt.figure(), None
            spectra = self._get_spectra_by_label_imname(cluster_label, image_basename, 100)
            for i in range(100):
                plt.plot(self.wavelengths, spectra[i, :])

            plt.xlabel('wavelength [nm]')
            plt.ylabel('%s %s\nreflectance [0-10000]' % (self.satellite, self.sensor))
            plt.title('Cluster #%s' % cluster_label)

            if include_mean_spectrum:
                plt.plot(self.wavelengths, np.mean(spectra, axis=0), c='black', lw=3)
            if include_median_spectrum:
                plt.plot(self.wavelengths, np.median(spectra, axis=0), '--', c='black', lw=3)

        # create a plot with multiple subplots
        else:
            nplots = len(lbls2plot)
            ncols = nplots if nplots < ncols else ncols
            nrows = nplots // ncols if not nplots % ncols else nplots // ncols + 1
            figsize = (4 * ncols, 3 * nrows)
            fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, sharex='all', sharey='all',
                                     **kw_fig)

            for lbl, ax in tqdm(zip(lbls2plot, axes.flatten()), total=nplots):
                spectra = self._get_spectra_by_label_imname(lbl, image_basename, 100)

                for i in range(100):
                    ax.plot(self.wavelengths, spectra[i, :], lw=1)

                if include_mean_spectrum:
                    ax.plot(self.wavelengths, np.mean(spectra, axis=0), c='black', lw=2)
                if include_median_spectrum:
                    ax.plot(self.wavelengths, np.median(spectra, axis=0), '--', c='black', lw=3)

                ax.grid(lw=0.2)
                ax.set_ylim(0, 10000)

                if ax.get_subplotspec().is_last_row():
                    ax.set_xlabel('wavelength [nm]')
                if ax.get_subplotspec().is_first_col():
                    ax.set_ylabel('%s %s\nreflectance [0-10000]' % (self.satellite, self.sensor))
                ax.set_title('Cluster #%s' % lbl)

        fig.suptitle("Refcube spectra from image '%s':" % image_basename, fontsize=15)
        plt.tight_layout(rect=(0, 0, 1, .95))
        plt.show()

        return fig



# -*- coding: utf-8 -*-

# spechomo, Spectral homogenization of multispectral satellite data
#
# Copyright (C) 2019-2021
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Please note the following exception: `spechomo` depends on tqdm, which is
# distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ClassifierNotAvailableError(RuntimeError):
    def __init__(self, spechomo_method, src_sat, src_sen, src_LBA, tgt_sat, tgt_sen, tgt_LBA, n_clusters):
        self.spechomo_method = spechomo_method
        self.src_sat = src_sat
        self.src_sen = src_sen
        self.src_LBA = src_LBA
        self.tgt_sat = tgt_sat
        self.tgt_sen = tgt_sen
        self.tgt_LBA = tgt_LBA
        self.n_clusters = n_clusters
        RuntimeError.__init__(self)

    def __str__(self):
        return 'No %s classifier available for predicting %s %s %s from %s %s %s (%d clusters).'\
               % (self.spechomo_method, self.tgt_sat, self.tgt_sen, self.tgt_LBA,
                  self.src_sat, self.src_sen, self.src_LBA, self.n_clusters)



# -*- coding: utf-8 -*-

# spechomo, Spectral homogenization of multispectral satellite data
#
# Copyright (C) 2019-2021
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Please note the following exception: `spechomo` depends on tqdm, which is
# distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .version import __version__, __versionalias__   # noqa (E402 + F401)
from .prediction import SpectralHomogenizer  # noqa (E402 + F401)


__author__ = """Daniel Scheffler"""
__email__ = 'daniel.scheffler@gfz-potsdam.de'



# -*- coding: utf-8 -*-

# spechomo, Spectral homogenization of multispectral satellite data
#
# Copyright (C) 2019-2021
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Please note the following exception: `spechomo` depends on tqdm, which is
# distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from glob import glob
import os
from zipfile import ZipFile
import dill
from multiprocessing import Pool
from typing import Union, List  # noqa F401  # flake8 issue
import json
import warnings
from urllib.request import urlretrieve, urlopen

import numpy as np  # noqa F401  # flake8 issue
from geoarray import GeoArray  # noqa F401  # flake8 issue
import pandas as pd
from tempfile import TemporaryDirectory
from natsort import natsorted
from tqdm import tqdm

from .options import options


def im2spectra(geoArr):
    # type: (Union[GeoArray, np.ndarray]) -> np.ndarray
    """Convert 3D images to array of spectra samples (rows: samples;  cols: spectral information)."""
    return geoArr.reshape((geoArr.shape[0] * geoArr.shape[1], geoArr.shape[2]))


def spectra2im(spectra, tgt_rows, tgt_cols):
    # type: (Union[GeoArray, np.ndarray], int, int) -> np.ndarray
    """Convert array of spectra samples (rows: samples;  cols: spectral information) to a 3D image.

    :param spectra:     2D array with rows: spectral samples / columns: spectral information (bands)
    :param tgt_rows:    number of target image rows
    :param tgt_cols:    number of target image rows
    :return:            3D array (rows x columns x spectral bands)
    """
    return spectra.reshape((tgt_rows, tgt_cols, spectra.shape[1]))


_columns_df_trafos = ['method', 'src_sat', 'src_sen', 'src_LBA', 'tgt_sat', 'tgt_sen', 'tgt_LBA', 'n_clusters']


def list_available_transformations(classifier_rootDir=options['classifiers']['rootdir'],
                                   method=None,
                                   src_sat=None, src_sen=None, src_LBA=None,
                                   tgt_sat=None, tgt_sen=None, tgt_LBA=None,
                                   n_clusters=None):
    # type: (str, str, str, str, List[str], str, str, List[str], int) -> pd.DataFrame
    """List all sensor transformations available according to the given classifier root directory.

    NOTE:   This function can be used to copy/paste possible input parameters for
            spechomo.SpectralHomogenizer.predict_by_machine_learner().

    :param classifier_rootDir:  directory containing classifiers for homogenization, either as .zip archives or
                                as .dill files
    :param method:              filter results by the machine learning approach to be used for spectral bands prediction
    :param src_sat:             filter results by source satellite, e.g., 'Landsat-8'
    :param src_sen:             filter results by source sensor, e.g., 'OLI_TIRS'
    :param src_LBA:             filter results by source bands list
    :param tgt_sat:             filter results by target satellite, e.g., 'Landsat-8'
    :param tgt_sen:             filter results by target sensor, e.g., 'OLI_TIRS'
    :param tgt_LBA:             filter results by target bands list
    :param n_clusters:          filter results by the number of spectral clusters to be used during LR/ RR/ QR
                                homogenization
    :return:    pandas.DataFrame listing all the available transformations
    """
    clf_zipfiles = glob(os.path.join(classifier_rootDir, '*_classifiers.zip'))

    df = pd.DataFrame(columns=_columns_df_trafos)

    if clf_zipfiles:
        for path_zipfile in clf_zipfiles:

            with TemporaryDirectory() as td, ZipFile(path_zipfile) as zF:
                zF.extractall(td)

                with Pool() as pool:
                    paths_dillfiles = [(os.path.join(td, fN)) for fN in natsorted(zF.namelist())]
                    dfs = pool.map(explore_classifer_dillfile, paths_dillfiles)

                    pool.close()  # needed to make coverage work in multiprocessing
                    pool.join()

                df = pd.concat([df] + dfs)

    else:
        paths_dillfiles = natsorted(glob(os.path.join(classifier_rootDir, '*.dill')))

        if paths_dillfiles:
            with Pool() as pool:
                dfs = pool.map(explore_classifer_dillfile, paths_dillfiles)
                pool.close()  # needed to make coverage work in multiprocessing
                pool.join()

            df = pd.concat([df] + dfs)

    # apply filters
    if not df.empty:
        df = df.reset_index(drop=True)

        def _force_list(arg):
            return [arg] if not isinstance(arg, list) else arg

        for filterparam in _columns_df_trafos:
            if locals()[filterparam]:
                df = df[df[filterparam].isin(_force_list(locals()[filterparam]))]

    return df


def explore_classifer_dillfile(path_dillFile):
    # type: (str) -> pd.DataFrame
    """List all homogenization transformations included in the given .dill file.

    :param path_dillFile:
    :return:
    """
    df = pd.DataFrame(columns=_columns_df_trafos)

    meth_nclust_str, src_sat, src_sen = os.path.splitext(os.path.basename(path_dillFile))[0].split('__')
    method, nclust_str = meth_nclust_str.split('_')
    nclust = int(nclust_str.split('clust')[1])

    with open(path_dillFile, 'rb') as inF:
        content = dill.load(inF)

        for src_LBA, subdict_L1 in content.items():
            for tgt_sat_sen, subdict_L2 in subdict_L1.items():
                tgt_sat, tgt_sen = tgt_sat_sen

                for tgt_LBA, subdict_L3 in subdict_L2.items():
                    df.loc[len(df)] = [method,
                                       src_sat, src_sen, src_LBA.split('__'),
                                       tgt_sat, tgt_sen, tgt_LBA.split('__'),
                                       nclust]

    return df


def export_classifiers_as_JSON(export_rootDir,
                               classifier_rootDir=options['classifiers']['rootdir'],
                               method=None,
                               src_sat=None, src_sen=None, src_LBA=None,
                               tgt_sat=None, tgt_sen=None, tgt_LBA=None,
                               n_clusters=None):
    # type: (str, str, str, str, str, List[str], str, str, List[str], int) -> None
    """Export spectral harmonization classifiers as JSON files that match the provided filtering criteria.

    NOTE: So far, this function will only work for LR classifiers.

    :param export_rootDir:      directory where to save the exported JSON files
    :param classifier_rootDir:  directory containing classifiers for homogenization, either as .zip archives or
                                as .dill files
    :param method:              filter by the machine learning approach to be used for spectral bands prediction
    :param src_sat:             filter by source satellite, e.g., 'Landsat-8'
    :param src_sen:             filter by source sensor, e.g., 'OLI_TIRS'
    :param src_LBA:             filter by source bands list
    :param tgt_sat:             filter by target satellite, e.g., 'Landsat-8'
    :param tgt_sen:             filter by target sensor, e.g., 'OLI_TIRS'
    :param tgt_LBA:             filter by target bands list
    :param n_clusters:          filter by the number of spectral clusters to be used during LR/ RR/ QR
                                homogenization
    :return:
    """
    from .classifier import Cluster_Learner

    # get matching classifiers
    df_trafos = list_available_transformations(classifier_rootDir=classifier_rootDir,
                                               method=method,
                                               src_sat=src_sat, src_sen=src_sen, src_LBA=src_LBA,
                                               tgt_sat=tgt_sat, tgt_sen=tgt_sen, tgt_LBA=tgt_LBA,
                                               n_clusters=n_clusters)

    if len(df_trafos):
        # export each matching transformation to a JSON file
        for i, trafo in tqdm(df_trafos.iterrows(), total=len(df_trafos)):

            CL = Cluster_Learner.from_disk(
                classifier_rootDir=classifier_rootDir,
                method=trafo.method,
                n_clusters=trafo.n_clusters,
                src_satellite=trafo.src_sat, src_sensor=trafo.src_sen, src_LBA=trafo.src_LBA,
                tgt_satellite=trafo.tgt_sat, tgt_sensor=trafo.tgt_sen, tgt_LBA=trafo.tgt_LBA)

            path_out = os.path.join(export_rootDir,
                                    trafo.method,
                                    "src__%s__%s__%s" % (trafo.src_sat, trafo.src_sen, '_'.join(trafo.src_LBA)),
                                    "to__%s__%s__%s" % (trafo.tgt_sat, trafo.tgt_sen, '_'.join(trafo.tgt_LBA)),
                                    "%s__clust%d.json" % (trafo.method, trafo.n_clusters)
                                    )
            os.makedirs(os.path.dirname(path_out), exist_ok=True)

            with open(path_out, 'w') as outF:
                json.dump(CL.to_jsonable_dict(), outF, indent=4, sort_keys=True)

    else:
        warnings.warn('No classifiers found matching the provided filter criteria. Nothing exported.', RuntimeWarning)


def download_pretrained_classifiers(method, tgt_dir=options['classifiers']['rootdir']):
    remote_filespecs = {
        '100k_conservrsp_SCA_SD100percSA90perc_without_aviris__SCADist90pSAM40p': {
            # 'LR': 'https://nextcloud.gfz-potsdam.de/s/Rzb75kckBreFfNE/download',  # 20201008
            'LR': 'https://nextcloud.gfz-potsdam.de/s/mZEnS5g7AGWyRHB/download',
            # 'QR': 'https://nextcloud.gfz-potsdam.de/s/Kk4zoCXxAEkAFZL/download',  # 20201008
            'QR': 'https://nextcloud.gfz-potsdam.de/s/JcQDbZBtTiw9NYi/download',
        }
    }
    clf_name = options['classifiers']['name']
    try:
        url = remote_filespecs[clf_name][method]
    except KeyError:
        raise RuntimeError("Currently there are no %s classifiers named '%s' available." % (method, clf_name))

    for i in range(3):  # try 3 times
        if i > 0:
            print('Download failed. Restarting...')

        # get filename
        fn = urlopen(url).headers['content-disposition'].split('filename=')[-1].replace('"', '')

        # download
        if not os.path.isdir(tgt_dir):
            os.makedirs(tgt_dir)
        outP, msg = urlretrieve(url, os.path.join(tgt_dir, fn))

        if os.path.getsize(outP) == int(msg.get('content-length')):
            return outP



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# spechomo, Spectral homogenization of multispectral satellite data
#
# Copyright (C) 2019-2021
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Please note the following exception: `spechomo` depends on tqdm, which is
# distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

version = {}
with open("spechomo/version.py") as version_file:
    exec(version_file.read(), version)

req = [
    'dill',
    'geoarray>=0.15.0',
    'matplotlib',
    'natsort',
    'nested_dict',
    'numpy',
    'pandas',
    'pyrsr>=0.5.0',
    'pyyaml',
    'scikit-learn>=0.23.2',
    'scipy',
    'seaborn',
    'specclassify>=0.3.0',
    'tabulate',
    'tqdm',
]

req_setup = ['setuptools-git']  # needed for package_data version-controlled by GIT

req_test = ['pytest', 'pytest-cov', 'pytest-reporter-html1', 'urlchecker']

req_doc = ['sphinx-argparse', 'sphinx_rtd_theme', 'sphinx-autodoc-typehints']

req_lint = ['flake8', 'pycodestyle', 'pydocstyle']

req_dev = req_setup + req_test + req_doc + req_lint

setup(
    author="Daniel Scheffler",
    author_email='daniel.scheffler@gfz-potsdam.de',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description="Spectral homogenization of multispectral satellite data.",
    extras_require={
        "doc": req_doc,
        "test": req_test,
        "lint": req_lint,
        "dev": req_dev
    },
    install_requires=req,
    license="Apache-2.0",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    package_data={'spechomo': ['options_default.yaml',
                               # 'resources/**/**/*'
                               ]},
    keywords=['spechomo', 'spectral homogenization', 'sensor fusion', 'remote sensing'],
    name='spechomo',
    packages=find_packages(include=['spechomo'], exclude=['tests']),
    setup_requires=req_setup,
    test_suite='tests',
    tests_require=req_test,
    url='https://git.gfz-potsdam.de/geomultisens/spechomo',
    version=version['__version__'],
    zip_safe=False,
)



# -*- coding: utf-8 -*-

# spechomo, Spectral homogenization of multispectral satellite data
#
# Copyright (C) 2019-2021
# - Daniel Scheffler (GFZ Potsdam, daniel.scheffler@gfz-potsdam.de)
# - Helmholtz Centre Potsdam - GFZ German Research Centre for Geosciences Potsdam,
#   Germany (https://www.gfz-potsdam.de/)
#
# This software was developed within the context of the GeoMultiSens project funded
# by the German Federal Ministry of Education and Research
# (project grant code: 01 IS 14 010 A-C).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Please note the following exception: `spechomo` depends on tqdm, which is
# distributed under the Mozilla Public Licence (MPL) v2.0 except for the files
# "tqdm/_tqdm.py", "setup.py", "README.rst", "MANIFEST.in" and ".gitignore".
# Details can be found here: https://github.com/tqdm/tqdm/blob/master/LICENCE.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from glob import glob
from multiprocessing import cpu_count
from typing import List, Tuple, Union, Dict, TYPE_CHECKING  # noqa F401  # flake8 issue
import logging  # noqa F401  # flake8 issue
from itertools import product

import dill
import numpy as np
from nested_dict import nested_dict
from pandas import DataFrame
from tqdm import tqdm
from geoarray import GeoArray
from pyrsr import RSR
from pyrsr.sensorspecs import get_LayerBandsAssignment

if TYPE_CHECKING:
    # avoid the sklearn imports on module level to get rid of static TLS ImportError
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.linear_model import LinearRegression, Ridge
    from sklearn.pipeline import Pipeline

from .clustering import KMeansRSImage
from .resampling import SpectralResampler
from .training_data import RefCube
from .logging import SpecHomo_Logger
from .options import options


class ReferenceCube_Generator(object):
    """Class for creating reference cube that are later used as training data for SpecHomo_Classifier."""

    def __init__(self, filelist_refs, tgt_sat_sen_list=None, dir_refcubes='', n_clusters=10, tgt_n_samples=1000,
                 v=False, logger=None, CPUs=None, dir_clf_dump=''):
        # type: (List[str], List[Tuple[str, str]], str, int, int, bool, logging.Logger, Union[None, int], str) -> None
        """Initialize ReferenceCube_Generator.

        :param filelist_refs:   list of (hyperspectral) reference images,
                                representing BOA reflectance, scaled between 0 and 10000
        :param tgt_sat_sen_list:    list satellite/sensor tuples containing those sensors for which the reference cube
                                    is to be computed, e.g. [('Landsat-8', 'OLI_TIRS',), ('Landsat-5', 'TM')]
        :param dir_refcubes:    output directory for the generated reference cube
        :param n_clusters:      number of clusters to be used for clustering the input images (KMeans)
        :param tgt_n_samples:   number o spectra to be collected from each input image
        :param v:               verbose mode
        :param logger:          instance of logging.Logger()
        :param CPUs:            number CPUs to use for computation
        :param dir_clf_dump:    directory where to store the serialized KMeans classifier
        """
        if not filelist_refs:
            raise ValueError('The given list of reference images is empty.')

        # args + kwargs
        self.ims_ref = [filelist_refs, ] if isinstance(filelist_refs, str) else filelist_refs
        self.tgt_sat_sen_list = tgt_sat_sen_list or [
            ('Landsat-8', 'OLI_TIRS'),
            ('Landsat-7', 'ETM+'),
            ('Landsat-5', 'TM'),
            ('Sentinel-2A', 'MSI'),
            # ('Terra', 'ASTER'),  # currently does not work
            ('SPOT-4', 'HRVIR1'),
            ('SPOT-4', 'HRVIR2'),
            ('SPOT-5', 'HRG1'),
            ('SPOT-5', 'HRG2'),
            ('RapidEye-5', 'MSI')
            ]
        self.dir_refcubes = os.path.abspath(dir_refcubes) if dir_refcubes else ''
        self.n_clusters = n_clusters
        self.tgt_n_samples = tgt_n_samples
        self.v = v
        self.logger = logger or SpecHomo_Logger(__name__)  # must be pickable
        self.CPUs = CPUs or cpu_count()
        self.dir_clf_dump = dir_clf_dump

        # privates
        self._refcubes = \
            {(sat, sen): RefCube(satellite=sat, sensor=sen,
                                 LayerBandsAssignment=self._get_tgt_LayerBandsAssignment(sat, sen))
             for sat, sen in self.tgt_sat_sen_list}

        if dir_refcubes and not os.path.isdir(self.dir_refcubes):
            os.makedirs(self.dir_refcubes)

    @property
    def refcubes(self):
        # type: () -> Dict[Tuple[str, str]: RefCube]
        """Return a dict holding instances of RefCube for each target satellite / sensor of self.tgt_sat_sen_list."""
        if not self._refcubes:

            # fill self._ref_cubes with GeoArray instances of already existing reference cubes read from disk
            if self.dir_refcubes:
                for path_refcube in glob(os.path.join(self.dir_refcubes, 'refcube__*.bsq')):
                    # TODO check if that really works
                    # check if current refcube path matches the target refcube specifications
                    identifier = re.search('refcube__(.*).bsq', os.path.basename(path_refcube)).group(1)
                    sat, sen, nclust_str, nsamp_str = identifier.split('__')  # type: str
                    nclust, nsamp = int(nclust_str.split('nclust')[1]), int(nsamp_str.split('nclust')[1])
                    correct_specs = all([(sat, sen) in self.tgt_sat_sen_list,
                                         nclust == self.n_clusters,
                                         nsamp == self.tgt_n_samples])

                    # import the existing ref cube if it matches the target refcube specs
                    if correct_specs:
                        self._refcubes[(sat, sen)] = \
                            RefCube(satellite=sat, sensor=sen, filepath=path_refcube,
                                    LayerBandsAssignment=self._get_tgt_LayerBandsAssignment(sat, sen))

        return self._refcubes

    @staticmethod
    def _get_tgt_LayerBandsAssignment(tgt_sat, tgt_sen):
        # type: (str, str) -> list
        """Get the LayerBandsAssignment for the specified target sensor.

        NOTE:   The returned bands list always contains all possible bands. Panchromatic bands or bands not available
                after atmospheric correction are removed. Further band selections are later done using np.take().
        NOTE: We don't use L1A here because the signal of the additional bands at L1A is not predictable by spectral
              harmonization as these bands are not driven by surface albedo but by atmospheric conditions (945, 1373nm).

        :param tgt_sat:     target satellite
        :param tgt_sen:     target sensor
        :return:
        """
        return get_LayerBandsAssignment(tgt_sat, tgt_sen,
                                        no_pan=False, no_thermal=True, sort_by_cwl=True, after_ac=True)

    @staticmethod
    def _get_tgt_RSR_object(tgt_sat, tgt_sen):
        # type: (str, str) -> RSR
        """Get an RSR instance containing the spectral response functions for the target sensor.

        :param tgt_sat:     target satellite
        :param tgt_sen:     target sensor
        :return:
        """
        return RSR(tgt_sat, tgt_sen, no_pan=False, no_thermal=True, after_ac=True)

    def generate_reference_cubes(self, fmt_out='ENVI', try_read_dumped_clf=True, sam_classassignment=False,
                                 max_distance='80%', max_angle=6, nmin_unique_spectra=50, alg_nodata='radical',
                                 progress=True):
        # type: (str, bool, bool, int, Union[int, float, str], Union[int, float, str], str, bool) -> ReferenceCube_Generator.refcubes  # noqa
        """Generate reference spectra from all hyperspectral input images.

        Workflow:
        1. Clustering/classification of hyperspectral images and selection of a given number of random signatures
           (a. Spectral downsamling to lower spectral resolution (speedup))
           b. KMeans clustering
           c. Selection of the same number of signatures from each cluster to avoid unequal amount of training data.
        2. Spectral resampling of the selected hyperspectral signatures (for each input image)
        3. Add resampled spectra to reference cubes for each target sensor and write cubes to disk

        :param fmt_out:             output format (GDAL driver code)
        :param try_read_dumped_clf: try to read a prediciouly serialized KMeans classifier from disk
                                    (massively speeds up the RefCube generation)
        :param sam_classassignment: False: use minimal euclidian distance to assign classes to cluster centers
                                    True: use the minimal spectral angle to assign classes to cluster centers
        :param max_distance:    spectra with a larger spectral distance than the given value will be excluded from
                                random sampling.
                                - if given as string like '20%', the maximum spectral distance is computed as 20%
                                percentile within each cluster
        :param max_angle:       spectra with a larger spectral angle than the given value will be excluded from
                                random sampling.
                                - if given as string like '20%', the maximum spectral angle is computed as 20%
                                percentile within each cluster
        :param nmin_unique_spectra:   in case a cluster has less than the given number,
                                      do not include it in the reference cube (default: 50)
        :param alg_nodata:      algorithm how to deal with pixels where the spectral bands of the source image
                                contain nodata within the spectral response of a target band
                                    'radical':      set output band to nodata
                                    'conservative': use existing spectral information and ignore nodata
                                                    (might alter the output spectral information,
                                                     e.g., at spectral absorption bands)
        :param progress:        show progress bar (default: True)
        :return:                np.array: [tgt_n_samples x images x spectral bands of the target sensor]
        """
        for im in self.ims_ref:  # type: Union[str, GeoArray]
            # TODO implement check if current image is already included in the refcube -> skip in that case
            src_im = GeoArray(im)

            # get random spectra of the original (hyperspectral) image, equally distributed over all computed clusters
            baseN = os.path.splitext(os.path.basename(im))[0] if isinstance(im, str) else im.basename
            unif_random_spectra = \
                self.cluster_image_and_get_uniform_spectra(src_im,
                                                           try_read_dumped_clf=try_read_dumped_clf,
                                                           sam_classassignment=sam_classassignment,
                                                           max_distance=max_distance,
                                                           max_angle=max_angle,
                                                           nmin_unique_spectra=nmin_unique_spectra,
                                                           progress=progress,
                                                           basename_clf_dump=baseN).astype(np.int16)

            # resample the set of random spectra to match the spectral characteristics of all target sensors
            for tgt_sat, tgt_sen in self.tgt_sat_sen_list:
                # perform spectral resampling
                self.logger.info('Performing spectral resampling to match %s %s specifications...' % (tgt_sat, tgt_sen))
                unif_random_spectra_rsp = self.resample_spectra(
                    unif_random_spectra,
                    src_cwl=np.array(src_im.meta.band_meta['wavelength'], dtype=float).flatten(),
                    tgt_rsr=self._get_tgt_RSR_object(tgt_sat, tgt_sen),
                    nodataVal=src_im.nodata,
                    alg_nodata=alg_nodata)

                # add the spectra as GeoArray instance to the in-mem ref cubes
                refcube = self.refcubes[(tgt_sat, tgt_sen)]  # type: RefCube
                refcube.add_spectra(unif_random_spectra_rsp, src_imname=src_im.basename,
                                    LayerBandsAssignment=self._get_tgt_LayerBandsAssignment(tgt_sat, tgt_sen))

                # update the reference cubes on disk
                if self.dir_refcubes:
                    refcube.save(path_out=os.path.join(self.dir_refcubes, 'refcube__%s__%s__nclust%s__nsamp%s.bsq'
                                                       % (tgt_sat, tgt_sen, self.n_clusters, self.tgt_n_samples)),
                                 fmt=fmt_out)

        return self.refcubes

    def cluster_image_and_get_uniform_spectra(self, im, downsamp_sat=None, downsamp_sen=None, basename_clf_dump='',
                                              try_read_dumped_clf=True, sam_classassignment=False, max_distance='80%',
                                              max_angle=6, nmin_unique_spectra=50, progress=False):
        # type: (Union[str, GeoArray, np.ndarray], str, str, str, bool, bool, int, Union[int, float, str], Union[int, float, str], bool) -> np.ndarray  # noqa
        """Compute KMeans clusters for the given image and return the an array of uniform random samples.

        :param im:              image to be clustered
        :param downsamp_sat:    satellite code used for intermediate image dimensionality reduction (input image is
                                spectrally resampled to this satellite before it is clustered). requires downsamp_sen.
                                If it is None, no intermediate downsampling is performed.
        :param downsamp_sen:    sensor code used for intermediate image dimensionality reduction (requires downsamp_sat)
        :param basename_clf_dump:   basename of serialized KMeans classifier
        :param try_read_dumped_clf: try to read a previously serialized KMeans classifier from disk
                                    (massively speeds up the RefCube generation)
        :param sam_classassignment: False: use minimal euclidian distance to assign classes to cluster centers
                                    True: use the minimal spectral angle to assign classes to cluster centers
        :param max_distance:    spectra with a larger spectral distance than the given value will be excluded from
                                random sampling.
                                - if given as string like '20%', the maximum spectral distance is computed as 20%
                                percentile within each cluster
        :param max_angle:       spectra with a larger spectral angle than the given value will be excluded from
                                random sampling.
                                - if given as string like '20%', the maximum spectral angle is computed as 20%
                                percentile within each cluster
        :param nmin_unique_spectra:   in case a cluster has less than the given number,
                                      do not include it in the reference cube (default: 50)
        :param progress:        whether to show progress bars or not
        :return:    2D array (rows: tgt_n_samples, columns: spectral information / bands
        """
        # input validation
        if downsamp_sat and not downsamp_sen or downsamp_sen and not downsamp_sat:
            raise ValueError("The parameters 'spec_rsp_sat' and 'spec_rsp_sen' must both be provided or completely "
                             "omitted.")

        im2clust = GeoArray(im)

        # get a KMeans classifier for the hyperspectral image
        path_clf = os.path.join(self.dir_clf_dump, '%s__KMeansClf.dill' % basename_clf_dump)
        if not sam_classassignment:
            path_clustermap = os.path.join(self.dir_clf_dump,
                                           '%s__KMeansClusterMap_nclust%d.bsq' % (basename_clf_dump, self.n_clusters))
        else:
            path_clustermap = os.path.join(self.dir_clf_dump,
                                           '%s__SAMClusterMap_nclust%d.bsq' % (basename_clf_dump, self.n_clusters))

        if try_read_dumped_clf and os.path.isfile(path_clf):
            # read the previously dumped classifier from disk
            self.logger.info('Reading previously serialized KMeans classifier from %s...' % path_clf)
            kmeans = KMeansRSImage.from_disk(path_clf=path_clf, im=im2clust)

        else:
            # create KMeans classifier
            # first, perform spectral resampling to Sentinel-2 to reduce dimensionality (speedup)
            if downsamp_sat and downsamp_sen:
                # NOTE: The KMeansRSImage class already reduces the input data to 1 million spectra by default
                tgt_rsr = RSR(satellite=downsamp_sat, sensor=downsamp_sen, no_pan=True, no_thermal=True)
                im2clust = self.resample_image_spectrally(im2clust, tgt_rsr, progress=progress)  # output = int16

            # compute KMeans clusters for the spectrally resampled image
            # NOTE: Nodata values are ignored during KMeans clustering.
            self.logger.info('Computing %s KMeans clusters from the input image %s...'
                             % (self.n_clusters, im2clust.basename))
            kmeans = KMeansRSImage(im2clust, n_clusters=self.n_clusters, sam_classassignment=sam_classassignment,
                                   CPUs=self.CPUs, v=self.v)
            kmeans.compute_clusters()
            kmeans.compute_spectral_distances()

            if self.dir_clf_dump and basename_clf_dump:
                kmeans.save_clustermap(path_clustermap)
                kmeans.dump(path_clf)

        if self.v:
            kmeans.plot_cluster_centers()
            kmeans.plot_cluster_histogram()

        # randomly grab the given number of spectra from each cluster, restricted to the 30 % purest spectra
        #   -> no spectra containing nodata values are returned
        self.logger.info('Getting %s random spectra from each cluster...' % (self.tgt_n_samples // self.n_clusters))
        random_samples = kmeans.get_random_spectra_from_each_cluster(samplesize=self.tgt_n_samples // self.n_clusters,
                                                                     max_distance=max_distance,
                                                                     max_angle=max_angle,
                                                                     nmin_unique_spectra=nmin_unique_spectra)
        # random_samples = kmeans\
        #     .get_purest_spectra_from_each_cluster(src_im=GeoArray(im),
        #                                           samplesize=self.tgt_n_samples // self.n_clusters)

        # combine the spectra (2D arrays) of all clusters to a single 2D array
        self.logger.info('Combining random samples from all clusters.')
        random_samples = np.vstack([random_samples[clusterlabel] for clusterlabel in random_samples])

        return random_samples

    def resample_spectra(self, spectra, src_cwl, tgt_rsr, nodataVal, alg_nodata='radical'):
        # type: (Union[GeoArray, np.ndarray], Union[list, np.array], RSR, int, str) -> np.ndarray
        """Perform spectral resampling of the given image to match the given spectral response functions.

        :param spectra:     2D array (rows: spectral samples;  columns: spectral information / bands
        :param src_cwl:     central wavelength positions of input spectra
        :param tgt_rsr:     target relative spectral response functions to be used for spectral resampling
        :param nodataVal:   nodata value of the given spectra to be ignored during resampling
        :param alg_nodata:  algorithm how to deal with pixels where the spectral bands of the source image
                            contain nodata within the spectral response of a target band
                            'radical':      set output band to nodata
                            'conservative': use existing spectral information and ignore nodata
                                            (might alter the outpur spectral information,
                                             e.g., at spectral absorption bands)
        :return:
        """
        spectra = GeoArray(spectra)

        # perform spectral resampling of input image to match spectral properties of target sensor
        self.logger.info('Performing spectral resampling to match spectral properties of %s %s...'
                         % (tgt_rsr.satellite, tgt_rsr.sensor))

        SR = SpectralResampler(src_cwl, tgt_rsr)
        spectra_rsp = SR.resample_spectra(spectra,
                                          chunksize=200,
                                          nodataVal=nodataVal,
                                          alg_nodata=alg_nodata,
                                          CPUs=self.CPUs)

        return spectra_rsp

    def resample_image_spectrally(self, src_im, tgt_rsr, src_nodata=None, alg_nodata='radical', progress=False):
        # type: (Union[str, GeoArray], RSR, Union[float, int], str, bool) -> Union[GeoArray, None]
        """Perform spectral resampling of the given image to match the given spectral response functions.

        :param src_im:      source image to be resampled
        :param tgt_rsr:     target relative spectral response functions to be used for spectral resampling
        :param src_nodata:  source image nodata value
        :param alg_nodata:  algorithm how to deal with pixels where the spectral bands of the source image
                            contain nodata within the spectral response of a target band
                            'radical':      set output band to nodata
                            'conservative': use existing spectral information and ignore nodata (might alter the output
                                            spectral information, e.g., at spectral absorption bands)
        :param progress:    show progress bar (default: false)
        :return:
        """
        # handle src_im provided as file path or GeoArray instance
        if isinstance(src_im, str):
            im_name = os.path.basename(src_im)
            im_gA = GeoArray(src_im, nodata=src_nodata)
        else:
            im_name = src_im.basename
            im_gA = src_im

        # read input image
        self.logger.info('Reading the input image %s...' % im_name)
        im_gA.cwl = np.array(im_gA.meta.band_meta['wavelength'], dtype=float).flatten()

        # perform spectral resampling of input image to match spectral properties of target sensor
        self.logger.info('Performing spectral resampling to match spectral properties of %s %s...'
                         % (tgt_rsr.satellite, tgt_rsr.sensor))
        SR = SpectralResampler(im_gA.cwl, tgt_rsr)

        tgt_im = GeoArray(np.zeros((*im_gA.shape[:2], len(tgt_rsr.bands)), dtype=np.int16),
                          geotransform=im_gA.gt,
                          projection=im_gA.prj,
                          nodata=im_gA.nodata)
        tgt_im.meta.band_meta['wavelength'] = list(tgt_rsr.wvl)
        tgt_im.bandnames = ['B%s' % i if len(i) == 2 else 'B0%s' % i for i in tgt_rsr.LayerBandsAssignment]

        tiles = im_gA.tiles((1000, 1000))  # use tiles to save memory
        for ((rS, rE), (cS, cE)), tiledata in (tqdm(tiles) if progress else tiles):
            tgt_im[rS: rE + 1, cS: cE + 1, :] = SR.resample_image(tiledata.astype(np.int16),
                                                                  nodataVal=im_gA.nodata,
                                                                  alg_nodata=alg_nodata,
                                                                  CPUs=self.CPUs)

        return tgt_im


class ClusterClassifier_Generator(object):
    """Class for creating collections of machine learning classifiers that can be used for spectral homogenization."""

    def __init__(self, list_refcubes, logger=None):
        # type: (List[Union[str, RefCube]], logging.Logger) -> None
        """Get an instance of Classifier_Generator.

        :param list_refcubes:   list of RefCube instances or paths for which the classifiers are to be created.
        :param logger:          instance of logging.Logger()
        """
        if not list_refcubes:
            raise ValueError('The given list of RefCube instances or paths is empty.')

        self.refcubes = [RefCube(inRC) if isinstance(inRC, str) else inRC for inRC in list_refcubes]
        self.logger = logger or SpecHomo_Logger(__name__)  # must be pickable

        # validation
        shapes = list(set([RC.data.shape[:2] for RC in self.refcubes]))
        if not len(shapes) == 1:
            raise ValueError('Unequal dimensions of the given reference cubes: %s' % shapes)

    @staticmethod
    def _get_derived_LayerBandsAssignments(satellite, sensor):
        """Get a list of possible LayerBandsAssignments in which the spectral training data may be arranged.

        :param satellite:   satellite to return LayerBandsAssignments for
        :param sensor:      sensor to return LayerBandsAssignments for
        :return:            e.g. for Landsat-8 OLI_TIRS:
                            [['1', '2', '3', '4', '5', '9', '6', '7'],
                             ['1', '2', '3', '4', '5', '9', '6', '7', '8'],
                             ['1', '2', '3', '4', '5', '6', '7'],
                             ['1', '2', '3', '4', '5', '6', '7'], ...]
        """
        # get bands (after AC)
        # NOTE: - signal of additional bands at L1A is not predictable by spectral harmonization because these bands
        #         are not driven by surface albedo but by atmospheric conditions (945, 1373 nm)

        LBAs = []
        kw = dict(satellite=satellite, sensor=sensor, no_thermal=True, after_ac=True)
        for lba in [get_LayerBandsAssignment(no_pan=False, sort_by_cwl=True, **kw),  # L1C_withPan_cwlSorted
                    get_LayerBandsAssignment(no_pan=True, sort_by_cwl=True, **kw),  # L1C_noPan_cwlSorted
                    get_LayerBandsAssignment(no_pan=False, sort_by_cwl=False, **kw),  # L1C_withPan_alphabetical
                    get_LayerBandsAssignment(no_pan=True, sort_by_cwl=False, **kw),  # L1C_noPan_alphabetical
                    ]:
            if lba not in LBAs:
                LBAs.append(lba)

        return LBAs

    @staticmethod
    def train_machine_learner(train_X, train_Y, test_X, test_Y, method, **kwargs):
        # type: (np.ndarray, np.ndarray, np.ndarray, np.ndarray, str, dict) -> Union[LinearRegression, Ridge, Pipeline, RandomForestRegressor]  # noqa E501 (line too long)
        """Use the given train and test data to train a machine learner and append some accuracy statistics.

        :param train_X:     reference training data
        :param train_Y:     target training data
        :param test_X:      reference test data
        :param test_Y:      target test data
        :param method:      type of machine learning classifiers to be included in classifier collections
                            'LR':   Linear Regression
                            'RR':   Ridge Regression
                            'QR':   Quadratic Regression
                            'RFR':  Random Forest Regression (50 trees)
        :param kwargs:      keyword arguments to be passed to the __init__() function of machine learners
        """
        from sklearn.pipeline import Pipeline

        ###################
        # train the model #
        ###################

        ML = get_machine_learner(method, **kwargs)
        ML.fit(train_X, train_Y)

        def mean_absolute_percentage_error(y_true, y_pred):
            y_true, y_pred = np.array(y_true), np.array(y_pred)

            # avoid division by 0
            if 0 in y_true:
                y_true = y_true.astype(float)
                y_true[y_true == 0] = np.nan

            return np.nanmean(np.abs((y_true - y_pred) / y_true), axis=0) * 100

        ###########################
        # compute RMSE, MAE, MAPE #
        ###########################

        from sklearn.metrics import mean_squared_error, mean_absolute_error
        predicted = ML.predict(test_X)  # returns 2D array (spectral samples x bands), e.g. 640 x 6
        # NOTE: 'raw_values': RMSE is column-wise computed
        #       => yields the same result as one would compute the RMSE band by band
        rmse = np.sqrt(mean_squared_error(test_Y, predicted, multioutput='raw_values')).astype(np.float32)
        mae = mean_absolute_error(test_Y, predicted, multioutput='raw_values').astype(np.float32)
        mape = mean_absolute_percentage_error(test_Y, predicted).astype(np.float32)

        # predicted_train = ML.predict(train_X)
        # rmse_train = np.sqrt(mean_squared_error(train_Y, predicted_train, multioutput='raw_values'))
        # # v2
        # test_Y_im = spectra2im(test_Y, tgt_rows=int(tgt_data.rows * 0.4), tgt_cols=tgt_data.cols)
        # pred_im = spectra2im(predicted, tgt_rows=int(tgt_data.rows * 0.4), tgt_cols=tgt_data.cols)
        # rmse_v2 = np.sqrt(mean_squared_error(test_Y_im[:,:,1], pred_im[:,:,1]))

        # append some metadata
        ML.scores = dict(train=ML.score(train_X, train_Y),
                         test=ML.score(test_X, test_Y))  # r2 scores
        ML.rmse_per_band = list(rmse)
        ML.mae_per_band = list(mae)
        ML.mape_per_band = list(mape)
        from .version import __version__, __versionalias__
        ML.spechomo_version = __version__
        ML.spechomo_versionalias = __versionalias__

        # convert float64 attributes to float32 to save memory (affects <0,05% of homogenized pixels by 1 DN)
        for attr in ['coef_', 'intercept_', 'singular_', '_residues']:
            if isinstance(ML, Pipeline):
                # noinspection PyProtectedMember
                setattr(ML._final_estimator, attr, getattr(ML._final_estimator, attr).astype(np.float32))
            else:
                try:
                    setattr(ML, attr, getattr(ML, attr).astype(np.float32))
                except AttributeError:
                    pass

        return ML

    def create_classifiers(self, outDir, method='LR', n_clusters=50, sam_classassignment=False, CPUs=None,
                           max_distance=options['classifiers']['trainspec_filtering']['max_distance'],
                           max_angle=options['classifiers']['trainspec_filtering']['max_angle'],
                           **kwargs):
        # type: (str, str, int, bool, int, Union[int, str], Union[int, str], dict) -> None
        """Create cluster classifiers for all combinations of the reference cubes given in __init__().

        :param outDir:      output directory for the created cluster classifier collections
        :param method:      type of machine learning classifiers to be included in classifier collections
                            'LR':   Linear Regression
                            'RR':   Ridge Regression
                            'QR':   Quadratic Regression
                            'RFR':  Random Forest Regression (50 trees with maximum depth of 3 by default)
        :param n_clusters:  number of clusters to be used for KMeans clustering
        :param sam_classassignment: False: use minimal euclidian distance to assign classes to cluster centers
                                    True: use the minimal spectral angle to assign classes to cluster centers
        :param CPUs:        number of CPUs to be used for KMeans clustering
        :param max_distance:    maximum spectral distance allowed during filtering of training spectra
                                - if given as string, e.g., '80%' excludes the worst 20 % of the input spectra
        :param max_angle:       maximum spectral angle allowed during filtering of training spectra
                                - if given as string, e.g., '80%' excludes the worst 20 % of the input spectra
        :param kwargs:      keyword arguments to be passed to machine learner
        """
        from sklearn.model_selection import train_test_split

        # validate and set defaults
        if not os.path.isdir(outDir):
            os.makedirs(outDir)

        if method == 'RFR':
            if n_clusters > 1:
                self.logger.warning("The spectral homogenization method 'Random Forest Regression' does not allow "
                                    "spectral sub-clustering. Setting 'n_clusters' to 1.")
                n_clusters = 1

            kwargs.update(dict(
                n_jobs=kwargs.get('n_jobs', CPUs),
                n_estimators=kwargs.get('n_estimators', options['classifiers']['RFR']['n_trees']),
                max_depth=kwargs.get('max_depth', options['classifiers']['RFR']['max_depth']),
                max_features=kwargs.get('max_features', options['classifiers']['RFR']['max_features'])
            ))

        # build the classifier collections with separate classifiers for each cluster
        for src_cube in self.refcubes:  # type: RefCube
            cls_collection = nested_dict()

            # get cluster labels for each source cube separately (as they depend on source spectral characteristics)
            clusterlabels_src_cube, spectral_distances, spectral_angles = \
                self._get_cluster_labels_for_source_refcube(src_cube, n_clusters, CPUs,
                                                            sam_classassignment=sam_classassignment,
                                                            return_spectral_distances=True,
                                                            return_spectral_angles=True)

            for tgt_cube in self.refcubes:
                if (src_cube.satellite, src_cube.sensor) == (tgt_cube.satellite, tgt_cube.sensor):
                    continue
                self.logger.info("Creating %s cluster classifier to predict %s %s from %s %s..."
                                 % (method, tgt_cube.satellite, tgt_cube.sensor, src_cube.satellite, src_cube.sensor))

                src_derived_LBAs = self._get_derived_LayerBandsAssignments(src_cube.satellite, src_cube.sensor)
                tgt_derived_LBAs = self._get_derived_LayerBandsAssignments(tgt_cube.satellite, tgt_cube.sensor)

                for src_LBA, tgt_LBA in product(src_derived_LBAs, tgt_derived_LBAs):
                    self.logger.debug('Creating %s cluster classifier for LBA %s => %s...'
                                      % (method, '_'.join(src_LBA), '_'.join(tgt_LBA)))

                    # Get center wavelength positions
                    # NOTE: they cannot be taken from RefCube instances because they always represent after-AC-LBAs
                    src_wavelengths = list(RSR(src_cube.satellite, src_cube.sensor, LayerBandsAssignment=src_LBA).wvl)
                    tgt_wavelengths = list(RSR(tgt_cube.satellite, tgt_cube.sensor, LayerBandsAssignment=tgt_LBA).wvl)

                    # Get training data for source and target image according to the given LayerBandsAssignments
                    # e.g., source: Landsat 7 image in LBA 1__2__3__4__5__7 and target L8 in 1__2__3__4__5__6__7
                    df_src_spectra_allclust = src_cube.get_spectra_dataframe(src_LBA)
                    df_tgt_spectra_allclust = tgt_cube.get_spectra_dataframe(tgt_LBA)

                    # assign clusterlabel and spectral distance to the cluster center to each spectrum
                    for df in [df_src_spectra_allclust, df_tgt_spectra_allclust]:
                        df.insert(0, 'cluster_label', clusterlabels_src_cube)
                        df.insert(1, 'spectral_distance', spectral_distances)
                        df.insert(2, 'spectral_angle', spectral_angles)

                    # remove spectra with cluster label -9999
                    # (clusters with too few spectra that are set to nodata in the refcube)
                    df_src_spectra_allclust = df_src_spectra_allclust[df_src_spectra_allclust.cluster_label != -9999]
                    df_tgt_spectra_allclust = df_tgt_spectra_allclust[df_tgt_spectra_allclust.cluster_label != -9999]

                    # remove spectra with -9999 in all bands that HAVE a valid clusterlabel
                    # NOTE: This may happen because the cluster labels are computed on the source cube only and may be
                    #       different on the target cube.
                    sub_bad = df_tgt_spectra_allclust[df_tgt_spectra_allclust.iloc[:, 3] == -9999].copy()
                    if not sub_bad.empty:
                        idx_bad = sub_bad[np.std(sub_bad.iloc[:, 3:], axis=1) == 0].index
                        if not idx_bad.empty:
                            df_src_spectra_allclust = df_src_spectra_allclust.drop(idx_bad)
                            df_tgt_spectra_allclust = df_tgt_spectra_allclust.drop(idx_bad)

                    # ensure source and target spectra do not contain nodata values (would affect classifiers)
                    assert src_cube.data.nodata is None or src_cube.data.nodata not in df_src_spectra_allclust.values
                    assert tgt_cube.data.nodata is None or tgt_cube.data.nodata not in df_tgt_spectra_allclust.values

                    for clusterlabel in range(n_clusters):
                        self.logger.debug('Creating %s classifier for cluster %s...' % (method, clusterlabel))

                        if method == 'RFR':
                            df_src_spectra_best = df_src_spectra_allclust.sample(10000, random_state=0, replace=True)
                            df_tgt_spectra_best = df_tgt_spectra_allclust.sample(10000, random_state=0, replace=True)
                        else:
                            if n_clusters == 1:
                                max_distance, max_angle = '100%', '100%'

                            df_src_spectra_best, df_tgt_spectra_best = \
                                self._extract_best_spectra_from_cluster(
                                    clusterlabel, df_src_spectra_allclust, df_tgt_spectra_allclust,
                                    max_distance=max_distance, max_angle=max_angle)

                        # Set train and test variables for the classifier
                        src_spectra_curlabel = df_src_spectra_best.values[:, 3:]
                        tgt_spectra_curlabel = df_tgt_spectra_best.values[:, 3:]

                        train_src, test_src, train_tgt, test_tgt = \
                            train_test_split(src_spectra_curlabel, tgt_spectra_curlabel,
                                             test_size=0.4, shuffle=True, random_state=0)

                        # train the learner and add metadata
                        ML = self.train_machine_learner(train_src, train_tgt, test_src, test_tgt, method, **kwargs)
                        # noinspection PyTypeChecker
                        ML = self._add_metadata_to_machine_learner(
                            ML, src_cube, tgt_cube, src_LBA, tgt_LBA, src_wavelengths, tgt_wavelengths,
                            train_src, train_tgt, n_clusters, clusterlabel)

                        # append to classifier collection
                        cls_collection[
                            '__'.join(src_LBA)][tgt_cube.satellite, tgt_cube.sensor][
                            '__'.join(tgt_LBA)][clusterlabel] = ML

            # dump to disk
            fName_cls = get_filename_classifier_collection(method, src_cube.satellite, src_cube.sensor,
                                                           n_clusters=n_clusters, **kwargs)

            with open(os.path.join(outDir, fName_cls), 'wb') as outF:
                dill.dump(cls_collection.to_dict(), outF,
                          protocol=dill.HIGHEST_PROTOCOL - 1)  # ensures some downwards compatibility

    def _get_cluster_labels_for_source_refcube(self, src_cube, n_clusters, CPUs, sam_classassignment=False,
                                               return_spectral_distances=False, return_spectral_angles=False):
        """Get cluster labels for each source cube separately.

        NOTE: - We use the GMS L1C bands (without atmospheric bands and PAN-band) for clustering.
              - clustering is only performed on the source cube because the source sensor spectral information
                is later used to assign the correct homogenization classifier to each pixel

        :param src_cube:
        :param n_clusters:
        :param sam_classassignment:         False: use minimal euclidian distance to assign classes to cluster centers
                                            True: use the minimal spectral angle to assign classes to cluster centers
        :param CPUs:
        :param return_spectral_distances:
        :return:
        """
        self.logger.info('Clustering %s %s reference cube (%s clusters)...'
                         % (src_cube.satellite, src_cube.sensor, n_clusters))
        LBA2clust = get_LayerBandsAssignment(src_cube.satellite, src_cube.sensor,
                                             no_pan=True, no_thermal=True, sort_by_cwl=True, after_ac=True)
        src_data2clust = src_cube.get_band_combination(LBA2clust)
        km = KMeansRSImage(src_data2clust, n_clusters=n_clusters, sam_classassignment=sam_classassignment, CPUs=CPUs)
        km.compute_clusters()

        return_vals = [km.labels_with_nodata]

        if return_spectral_distances:
            return_vals.append(km.spectral_distances_with_nodata)
        if return_spectral_angles:
            return_vals.append(km.spectral_angles_with_nodata)

        return tuple(return_vals)

    def _extract_best_spectra_from_cluster(self, clusterlabel, df_src_spectra_allclust, df_tgt_spectra_allclust,
                                           max_distance, max_angle):
        # NOTE: We exclude the noisy spectra with the largest spectral distances to their cluster
        #       center here (random spectra from within the upper 40 %)
        assert len(df_src_spectra_allclust.index) == len(df_tgt_spectra_allclust.index), \
            'Source and target spectra dataframes must have the same number of spectral samples.'

        df_src_spectra = df_src_spectra_allclust[df_src_spectra_allclust.cluster_label == clusterlabel]

        # max_dist = np.percentile(df_src_spectra.spectral_distance, max_distance)
        # max_angle = np.percentile(df_src_spectra.spectral_angle, max_angle_percent)

        if isinstance(max_angle, str):
            max_angle = np.percentile(df_src_spectra.spectral_angle,
                                      int(max_angle.split('%')[0].strip()))

        tmp = df_src_spectra[df_src_spectra.spectral_angle <= max_angle]
        if len(tmp.index) > 10:
            df_src_spectra = tmp
        else:
            df_src_spectra = df_src_spectra.sort_values(by='spectral_angle').head(10)
            self.logger.warning('Had to choose spectra with SA up to %.1f degrees for cluster #%s.'
                                % (np.max(df_src_spectra['spectral_angle']), clusterlabel))

        if isinstance(max_distance, str):
            max_distance = np.percentile(df_src_spectra.spectral_distance,
                                         int(max_distance.split('%')[0].strip()))

        tmp = df_src_spectra[df_src_spectra.spectral_distance <= max_distance]
        if len(tmp.index) > 10:
            df_src_spectra = tmp
        else:
            df_src_spectra = df_src_spectra.sort_values(by='spectral_distance').head(10)
            self.logger.warning('Had to choose spectra with SD up to %.1f for cluster #%s.'
                                % (np.max(df_src_spectra['spectral_distance']), clusterlabel))

        if len(df_src_spectra.index) > 1700:
            df_src_spectra = df_src_spectra.sort_values(by='spectral_angle').head(1700)

        df_tgt_spectra = df_tgt_spectra_allclust.loc[df_src_spectra.index, :]

        return df_src_spectra, df_tgt_spectra

    @staticmethod
    def _add_metadata_to_machine_learner(ML, src_cube, tgt_cube, src_LBA, tgt_LBA, src_wavelengths, tgt_wavelengths,
                                         src_train_spectra, tgt_train_spectra, n_clusters, clusterlabel):
        # compute cluster center for source spectra (only on those spectra used for model training)
        cluster_center = np.mean(src_train_spectra, axis=0)
        cluster_median = np.median(src_train_spectra, axis=0)
        sample_spectra = DataFrame(src_train_spectra).sample(100, replace=True, random_state=20).values

        ML.src_satellite = src_cube.satellite
        ML.tgt_satellite = tgt_cube.satellite
        ML.src_sensor = src_cube.sensor
        ML.tgt_sensor = tgt_cube.sensor
        ML.src_LBA = src_LBA
        ML.tgt_LBA = tgt_LBA
        ML.src_n_bands = len(ML.src_LBA)
        ML.tgt_n_bands = len(ML.tgt_LBA)
        ML.src_wavelengths = list(np.array(src_wavelengths).astype(np.float32))
        ML.tgt_wavelengths = list(np.array(tgt_wavelengths).astype(np.float32))
        ML.n_clusters = n_clusters
        ML.clusterlabel = clusterlabel
        ML.cluster_center = cluster_center.astype(src_cube.data.dtype)
        ML.cluster_median = cluster_median.astype(src_cube.data.dtype)
        ML.cluster_sample_spectra = sample_spectra.astype(np.int16)  # scaled between 0 and 10000
        ML.tgt_cluster_center = np.mean(tgt_train_spectra, axis=0).astype(tgt_cube.data.dtype)
        ML.tgt_cluster_median = np.median(tgt_train_spectra, axis=0).astype(tgt_cube.data.dtype)

        assert len(ML.src_LBA) == len(ML.src_wavelengths)
        assert len(ML.tgt_LBA) == len(ML.tgt_wavelengths)

        return ML


def get_machine_learner(method='LR', **init_params):
    # type: (str, dict) -> Union[LinearRegression, Ridge, Pipeline]
    """Get an instance of a machine learner.

    :param method:          'LR':   Linear Regression
                            'RR':   Ridge Regression
                            'QR':   Quadratic Regression
                            'RFR':  Random Forest Regression (50 trees)
    :param init_params:     parameters to be passed to __init__() function of the returned machine learner model.
    """
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.linear_model import LinearRegression, Ridge
    from sklearn.pipeline import make_pipeline
    from sklearn.preprocessing import PolynomialFeatures

    if method == 'LR':
        return LinearRegression(**init_params)
    elif method == 'RR':
        return Ridge(**init_params)
    elif method == 'QR':
        return make_pipeline(PolynomialFeatures(degree=2), LinearRegression(**init_params))
    elif method == 'RFR':
        return RandomForestRegressor(**init_params)
    else:
        raise ValueError("Unknown machine learner method code '%s'." % method)


def get_filename_classifier_collection(method, src_satellite, src_sensor, n_clusters=1, **cls_kwinit):
    if method == 'RR':
        method += '_alpha1.0'  # TODO add to config
    elif method == 'RFR':
        assert 'n_estimators' in cls_kwinit
        method += '_trees%s' % cls_kwinit['n_estimators']

    return '__'.join(['%s_clust%s' % (method, n_clusters), src_satellite, src_sensor]) + '.dill'


# def get_classifier_filename(method, src_satellite, src_sensor, src_LBA_name, tgt_satellite, tgt_sensor, tgt_LBA_name):
#     return '__'.join([method, src_satellite, src_sensor, src_LBA_name]) + \
#                     '__to__' + '__'.join([tgt_satellite, tgt_sensor, tgt_LBA_name]) + '.dill'
